# Copyright (c) 2009-2016 testtools developers. See LICENSE for details.

__all__ = ["IsDeprecated", "WarningMessage", "Warnings"]

import warnings
from collections.abc import Callable
from typing import Any

from ._basic import Is
from ._const import Always
from ._datastructures import MatchesListwise, MatchesStructure
from ._higherorder import (
    AfterPreprocessing,
    Annotate,
)
from ._impl import Matcher, Mismatch


def WarningMessage(
    category_type: type[Warning],
    message: "Matcher[Any] | None" = None,
    filename: "Matcher[Any] | None" = None,
    lineno: "Matcher[Any] | None" = None,
    line: "Matcher[Any] | None" = None,
) -> "MatchesStructure[warnings.WarningMessage]":
    r"""Create a matcher that will match `warnings.WarningMessage`\s.

    For example, to match captured `DeprecationWarning`s with a message about
    some ``foo`` being replaced with ``bar``:

    .. code-block:: python

       WarningMessage(DeprecationWarning,
                      message=MatchesAll(
                          Contains('foo is deprecated'),
                          Contains('use bar instead')))

    :param type category_type: A warning type, for example `DeprecationWarning`.
    :param message_matcher: A matcher object that will be evaluated against
        warning's message.
    :param filename_matcher: A matcher object that will be evaluated against
        the warning's filename.
    :param lineno_matcher: A matcher object that will be evaluated against the
        warning's line number.
    :param line_matcher: A matcher object that will be evaluated against the
        warning's line of source code.
    """
    category_matcher = Is(category_type)
    message_matcher: Matcher[Any] = message or Always()
    filename_matcher: Matcher[Any] = filename or Always()
    lineno_matcher: Matcher[Any] = lineno or Always()
    line_matcher: Matcher[Any] = line or Always()
    return MatchesStructure(
        category=Annotate("Warning's category type does not match", category_matcher),
        message=Annotate(
            "Warning's message does not match", AfterPreprocessing(str, message_matcher)
        ),
        filename=Annotate("Warning's filname does not match", filename_matcher),
        lineno=Annotate("Warning's line number does not match", lineno_matcher),
        line=Annotate("Warning's source line does not match", line_matcher),
    )


class Warnings(Matcher[Callable[[], Any]]):
    """Match if the matchee produces warnings."""

    def __init__(self, warnings_matcher: "Matcher[Any] | None" = None) -> None:
        """Create a Warnings matcher.

        :param warnings_matcher: Optional validator for the warnings emitted by
        matchee. If no warnings_matcher is supplied then the simple fact that
        at least one warning is emitted is considered enough to match on.
        """
        self.warnings_matcher = warnings_matcher

    def match(self, matchee: Callable[[], Any]) -> Mismatch | None:
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # Handle staticmethod objects by extracting the underlying function
            if isinstance(matchee, staticmethod):
                matchee = matchee.__func__
            matchee()
            if self.warnings_matcher is not None:
                return self.warnings_matcher.match(w)
            elif not w:
                return Mismatch("Expected at least one warning, got none")
        return None

    def __str__(self) -> str:
        return f"Warnings({self.warnings_matcher!s})"


def IsDeprecated(message: "Matcher[Any]") -> Warnings:
    """Make a matcher that checks that a callable produces exactly one
    `DeprecationWarning`.

    :param message: Matcher for the warning message.
    """
    return Warnings(
        MatchesListwise(
            [WarningMessage(category_type=DeprecationWarning, message=message)]
        )
    )



# Copyright (c) 2010-2016 testtools developers. See LICENSE for details.

"""Individual test case execution for tests that return Deferreds.

Example::

    class TwistedTests(testtools.TestCase):

        run_tests_with = AsynchronousDeferredRunTest

        def test_something(self):
            # Wait for 5 seconds and then fire with 'Foo'.
            d = Deferred()
            reactor.callLater(5, lambda: d.callback('Foo'))
            d.addCallback(self.assertEqual, 'Foo')
            return d

When ``test_something`` is run, ``AsynchronousDeferredRunTest`` will run the
reactor until ``d`` fires, and wait for all of its callbacks to be processed.
"""

__all__ = [
    "AsynchronousDeferredRunTest",
    "AsynchronousDeferredRunTestForBrokenTwisted",
    "SynchronousDeferredRunTest",
    "assert_fails_with",
]

import io
import sys
from typing import Any

from fixtures import CompoundFixture, Fixture
from twisted.internet import defer

from testtools.content import Content, text_content
from testtools.content_type import UTF8_TEXT
from testtools.runtest import RunTest, _raise_force_fail_error

from ._deferred import extract_result
from ._spinner import (
    NoResultError,
    Spinner,
    TimeoutError,
    trap_unhandled_errors,
)

try:
    from twisted.logger import globalLogPublisher
except ImportError:
    globalLogPublisher = None  # type: ignore[assignment]
from twisted.python import log

try:
    from twisted.trial.unittest import _LogObserver  # type: ignore[attr-defined]
except ImportError:
    from twisted.trial._synctest import _LogObserver


class _DeferredRunTest(RunTest):
    """Base for tests that return Deferreds."""

    def _got_user_failure(self, failure, tb_label="traceback"):
        """We got a failure from user code."""
        return self._got_user_exception(
            (failure.type, failure.value, failure.getTracebackObject()),
            tb_label=tb_label,
        )


class SynchronousDeferredRunTest(_DeferredRunTest):
    """Runner for tests that return synchronous Deferreds.

    This runner doesn't touch the reactor at all. It assumes that tests return
    Deferreds that have already fired.
    """

    def _run_user(self, function, *args):
        d = defer.maybeDeferred(function, *args)
        d.addErrback(self._got_user_failure)
        result = extract_result(d)
        return result


def _get_global_publisher_and_observers():
    """Return ``(log_publisher, observers)``.

    Twisted 15.2.0 changed the logging framework. This method will always
    return a tuple of the global log publisher and all observers associated
    with that publisher.
    """
    if globalLogPublisher is not None:
        # Twisted >= 15.2.0, with the new twisted.logger framework.
        # log.theLogPublisher.observers will only contain legacy observers;
        # we need to look at globalLogPublisher._observers, which contains
        # both legacy and modern observers, and add and remove them via
        # globalLogPublisher.  However, we must still add and remove the
        # observers we want to run with via log.theLogPublisher, because
        # _LogObserver may consider old keys and require them to be mapped.
        publisher = globalLogPublisher
        return (publisher, list(publisher._observers))
    else:
        publisher = log.theLogPublisher
        return (publisher, list(publisher.observers))


class _NoTwistedLogObservers(Fixture):
    """Completely but temporarily remove all Twisted log observers."""

    def _setUp(self):
        publisher, real_observers = _get_global_publisher_and_observers()
        for observer in reversed(real_observers):
            publisher.removeObserver(observer)
            self.addCleanup(publisher.addObserver, observer)


class _TwistedLogObservers(Fixture):
    """Temporarily add Twisted log observers."""

    def __init__(self, observers):
        super().__init__()
        self._observers = observers
        self._log_publisher = log.theLogPublisher

    def _setUp(self):
        for observer in self._observers:
            self._log_publisher.addObserver(observer)
            self.addCleanup(self._log_publisher.removeObserver, observer)


class _ErrorObserver(Fixture):
    """Capture errors logged while fixture is active."""

    def __init__(self, error_observer):
        super().__init__()
        self._error_observer = error_observer

    def _setUp(self):
        self.useFixture(_TwistedLogObservers([self._error_observer.gotEvent]))

    def flush_logged_errors(self, *error_types):
        """Clear errors of the given types from the logs.

        If no errors provided, clear all errors.

        :return: An iterable of errors removed from the logs.
        """
        return self._error_observer.flushErrors(*error_types)


class CaptureTwistedLogs(Fixture):
    """Capture all the Twisted logs and add them as a detail.

    Much of the time, you won't need to use this directly, as
    :py:class:`AsynchronousDeferredRunTest` captures Twisted logs when the
    ``store_twisted_logs`` is set to ``True`` (which it is by default).

    However, if you want to do custom processing of Twisted's logs, then this
    class can be useful.

    For example::

        class TwistedTests(TestCase):
            run_tests_with(
                partial(AsynchronousDeferredRunTest, store_twisted_logs=False))

            def setUp(self):
                super(TwistedTests, self).setUp()
                twisted_logs = self.useFixture(CaptureTwistedLogs())
                # ... do something with twisted_logs ...
    """

    LOG_DETAIL_NAME = "twisted-log"

    def _setUp(self):
        logs = io.StringIO()
        full_observer = log.FileLogObserver(logs)
        self.useFixture(_TwistedLogObservers([full_observer.emit]))
        self.addDetail(
            self.LOG_DETAIL_NAME,
            Content(UTF8_TEXT, lambda: [logs.getvalue().encode("utf-8")]),
        )


# Observer of the Twisted log that we install during tests.
#
# This is a global so that users can call flush_logged_errors errors in their
# test cases.
_log_observer = _LogObserver()


def flush_logged_errors(*error_types):
    """Flush errors of the given types from the global Twisted log.

    Any errors logged during a test will be bubbled up to the test result,
    marking the test as erroring. Use this function to declare that logged
    errors were expected behavior.

    For example::

        try:
            1/0
        except ZeroDivisionError:
            log.err()
        # Prevent logged ZeroDivisionError from failing the test.
        flush_logged_errors(ZeroDivisionError)

    :param error_types: A variable argument list of exception types.
    """
    # XXX: jml: I would like to deprecate this in favour of
    # _ErrorObserver.flush_logged_errors so that I can avoid mutable global
    # state. However, I don't know how to make the correct instance of
    # _ErrorObserver.flush_logged_errors available to the end user. I also
    # don't yet have a clear deprecation/migration path.
    return _log_observer.flushErrors(*error_types)


class AsynchronousDeferredRunTest(_DeferredRunTest):
    """Runner for tests that return Deferreds that fire asynchronously.

    Use this runner when you have tests that return Deferreds that will
    only fire if the reactor is left to spin for a while.
    """

    def __init__(
        self,
        case,
        handlers=None,
        last_resort=None,
        reactor=None,
        timeout=0.005,
        debug=False,
        suppress_twisted_logging=True,
        store_twisted_logs=True,
    ):
        """Construct an ``AsynchronousDeferredRunTest``.

        Please be sure to always use keyword syntax, not positional, as the
        base class may add arguments in future - and for core code
        compatibility with that we have to insert them before the local
        parameters.

        :param TestCase case: The `TestCase` to run.
        :param handlers: A list of exception handlers (ExceptionType, handler)
            where 'handler' is a callable that takes a `TestCase`, a
            ``testtools.TestResult`` and the exception raised.
        :param last_resort: Handler to call before re-raising uncatchable
            exceptions (those for which there is no handler).
        :param reactor: The Twisted reactor to use.  If not given, we use the
            default reactor.
        :param float timeout: The maximum time allowed for running a test.  The
            default is 0.005s.
        :param debug: Whether or not to enable Twisted's debugging.  Use this
            to get information about unhandled Deferreds and left-over
            DelayedCalls.  Defaults to False.
        :param bool suppress_twisted_logging: If True, then suppress Twisted's
            default logging while the test is being run. Defaults to True.
        :param bool store_twisted_logs: If True, then store the Twisted logs
            that took place during the run as the 'twisted-log' detail.
            Defaults to True.
        """
        super().__init__(case, handlers, last_resort)
        if reactor is None:
            from twisted.internet import reactor
        self._reactor = reactor
        self._timeout = timeout
        self._debug = debug
        self._suppress_twisted_logging = suppress_twisted_logging
        self._store_twisted_logs = store_twisted_logs

    @classmethod
    def make_factory(
        cls,
        reactor=None,
        timeout=0.005,
        debug=False,
        suppress_twisted_logging=True,
        store_twisted_logs=True,
    ):
        """Make a factory that conforms to the RunTest factory interface.

        Example::

            class SomeTests(TestCase):
                # Timeout tests after two minutes.
                run_tests_with = AsynchronousDeferredRunTest.make_factory(
                    timeout=120)
        """

        # This is horrible, but it means that the return value of the method
        # will be able to be assigned to a class variable *and* also be
        # invoked directly.
        class AsynchronousDeferredRunTestFactory:
            def __call__(self, case, handlers=None, last_resort=None):
                return cls(
                    case,
                    handlers,
                    last_resort,
                    reactor,
                    timeout,
                    debug,
                    suppress_twisted_logging,
                    store_twisted_logs,
                )

        return AsynchronousDeferredRunTestFactory()

    @defer.inlineCallbacks
    def _run_cleanups(self, result=None):
        """Run the cleanups on the test case.

        We expect that the cleanups on the test case can also return
        asynchronous Deferreds.  As such, we take the responsibility for
        running the cleanups, rather than letting TestCase do it.
        """
        last_exception = None
        while self.case._cleanups:
            f, args, kwargs = self.case._cleanups.pop()
            d = defer.maybeDeferred(f, *args, **kwargs)
            try:
                yield d
            except Exception:
                exc_info = sys.exc_info()
                self.case._report_traceback(exc_info)
                last_exception = exc_info[1]
        return last_exception

    def _make_spinner(self):
        """Make the `Spinner` to be used to run the tests."""
        return Spinner(self._reactor, debug=self._debug)

    def _run_deferred(self):
        """Run the test, assuming everything in it is Deferred-returning.

        This should return a Deferred that fires with True if the test was
        successful and False if the test was not successful.  It should *not*
        call addSuccess on the result, because there's reactor clean up that
        we needs to be done afterwards.
        """
        fails: list[Any] = []

        def fail_if_exception_caught(exception_caught):
            if self.exception_caught == exception_caught:
                fails.append(None)

        def clean_up(ignored=None):
            """Run the cleanups."""
            d = self._run_cleanups()

            def clean_up_done(result):
                if result is not None:
                    self._exceptions.append(result)
                    fails.append(None)

            return d.addCallback(clean_up_done)

        def set_up_done(exception_caught):
            """Set up is done, either clean up or run the test."""
            if self.exception_caught == exception_caught:
                fails.append(None)
                return clean_up()
            else:
                d = self._run_user(self.case._run_test_method, self.result)
                d.addCallback(fail_if_exception_caught)
                d.addBoth(tear_down)
                return d

        def tear_down(ignored):
            d = self._run_user(self.case._run_teardown, self.result)
            d.addCallback(fail_if_exception_caught)
            d.addBoth(clean_up)
            return d

        def force_failure(ignored):
            if getattr(self.case, "force_failure", None):
                d = self._run_user(_raise_force_fail_error)
                d.addCallback(fails.append)
                return d

        d = self._run_user(self.case._run_setup, self.result)
        d.addCallback(set_up_done)
        d.addBoth(force_failure)
        d.addBoth(lambda ignored: len(fails) == 0)
        return d

    def _log_user_exception(self, e):
        """Raise 'e' and report it as a user exception."""
        try:
            raise e
        except e.__class__:
            exc_info = sys.exc_info()
            assert exc_info[0] is not None and exc_info[1] is not None
            self._got_user_exception(exc_info)

    def _blocking_run_deferred(self, spinner):
        try:
            return trap_unhandled_errors(spinner.run, self._timeout, self._run_deferred)
        except NoResultError:
            # We didn't get a result at all!  This could be for any number of
            # reasons, but most likely someone hit Ctrl-C during the test.
            exc_info = sys.exc_info()
            assert exc_info[0] is not None and exc_info[1] is not None
            self._got_user_exception(exc_info)
            self.result.stop()
            return False, []
        except TimeoutError:
            # The function took too long to run.
            self._log_user_exception(TimeoutError(self.case, self._timeout))
            return False, []

    def _get_log_fixture(self):
        """Return the log fixture we're configured to use."""
        fixtures: list[Fixture] = []
        # TODO: Expose these fixtures and deprecate both of these options in
        # favour of them.
        if self._suppress_twisted_logging:
            fixtures.append(_NoTwistedLogObservers())
        if self._store_twisted_logs:
            fixtures.append(CaptureTwistedLogs())
        return CompoundFixture(fixtures)

    def _run_core(self):
        # XXX: Blatting over the namespace of the test case isn't a nice thing
        # to do. Find a better way of communicating between runtest and test
        # case.
        setattr(self.case, "reactor", self._reactor)
        spinner = self._make_spinner()

        # We can't just install these as fixtures on self.case, because we
        # need the clean up to run even if the test times out.
        #
        # See https://bugs.launchpad.net/testtools/+bug/897196.
        with self._get_log_fixture() as capture_logs:
            for name, detail in capture_logs.getDetails().items():
                self.case.addDetail(name, detail)
            with _ErrorObserver(_log_observer) as error_fixture:
                successful, unhandled = self._blocking_run_deferred(spinner)
            for logged_error in error_fixture.flush_logged_errors():
                successful = False
                self._got_user_failure(logged_error, tb_label="logged-error")

        if unhandled:
            successful = False
            for debug_info in unhandled:
                f = debug_info.failResult
                info = debug_info._getDebugTracebacks()
                if info:
                    self.case.addDetail(
                        "unhandled-error-in-deferred-debug", text_content(info)
                    )
                self._got_user_failure(f, "unhandled-error-in-deferred")

        junk = spinner.clear_junk()
        if junk:
            successful = False
            self._log_user_exception(UncleanReactorError(junk))

        if successful:
            self.result.addSuccess(self.case, details=self.case.getDetails())

    def _run_user(self, function, *args):
        """Run a user-supplied function.

        This just makes sure that it returns a Deferred, regardless of how the
        user wrote it.
        """
        d = defer.maybeDeferred(function, *args)
        return d.addErrback(self._got_user_failure)


class AsynchronousDeferredRunTestForBrokenTwisted(AsynchronousDeferredRunTest):
    """Test runner that works around Twisted brokenness re reactor junk.

    There are many APIs within Twisted itself where a Deferred fires but
    leaves cleanup work scheduled for the reactor to do.  Arguably, many of
    these are bugs.  This runner iterates the reactor event loop a number of
    times after every test, in order to shake out these buggy-but-commonplace
    events.
    """

    def _make_spinner(self):
        spinner = super()._make_spinner()
        spinner._OBLIGATORY_REACTOR_ITERATIONS = 2
        return spinner


def assert_fails_with(d, *exc_types, **kwargs):
    """Assert that ``d`` will fail with one of ``exc_types``.

    The normal way to use this is to return the result of
    ``assert_fails_with`` from your unit test.

    Equivalent to Twisted's ``assertFailure``.

    :param Deferred d: A ``Deferred`` that is expected to fail.
    :param exc_types: The exception types that the Deferred is expected to
        fail with.
    :param type failureException: An optional keyword argument.  If provided,
        will raise that exception instead of
        ``testtools.TestCase.failureException``.
    :return: A ``Deferred`` that will fail with an ``AssertionError`` if ``d``
        does not fail with one of the exception types.
    """
    failureException = kwargs.pop("failureException", None)
    if failureException is None:
        # Avoid circular imports.
        from testtools import TestCase

        failureException = TestCase.failureException
    expected_names = ", ".join(exc_type.__name__ for exc_type in exc_types)

    def got_success(result):
        raise failureException(f"{expected_names} not raised ({result!r} returned)")

    def got_failure(failure):
        if failure.check(*exc_types):
            return failure.value
        raise failureException(
            f"{failure.type.__name__} raised instead of {expected_names}:\n"
            f" {failure.getTraceback()}"
        )

    return d.addCallbacks(got_success, got_failure)


class UncleanReactorError(Exception):
    """Raised when the reactor has junk in it."""

    def __init__(self, junk):
        Exception.__init__(
            self,
            "The reactor still thinks it needs to do things. Close all "
            "connections, kill all processes and make sure all delayed "
            "calls have either fired or been cancelled:\n{}".format(
                "".join(map(self._get_junk_info, junk))
            ),
        )

    def _get_junk_info(self, junk):
        from twisted.internet.base import DelayedCall

        if isinstance(junk, DelayedCall):
            ret = str(junk)
        else:
            ret = repr(junk)
        return f"  {ret}\n"



# Copyright (c) testtools developers. See LICENSE for details.

"""Matchers that operate on synchronous Deferreds.

A "synchronous" Deferred is one that does not need the reactor or any other
asynchronous process in order to fire.

Normal application code can't know when a Deferred is going to fire, because
that is generally left up to the reactor. Unit tests can (and should!) provide
fake reactors, or don't use the reactor at all, so that Deferreds fire
synchronously.

These matchers allow you to make assertions about when and how Deferreds fire,
and about what values they fire with.
"""

from typing import Any

from testtools.matchers import Matcher, Mismatch

from ._deferred import failure_content, on_deferred_result


class _NoResult(Matcher[Any]):
    """Matches a Deferred that has not yet fired."""

    @staticmethod
    def _got_result(deferred, result):
        return Mismatch(f"No result expected on {deferred!r}, found {result!r} instead")

    def match(self, deferred):
        """Match ``deferred`` if it hasn't fired."""
        return on_deferred_result(
            deferred,
            on_success=self._got_result,
            on_failure=self._got_result,
            on_no_result=lambda _: None,
        )


_NO_RESULT = _NoResult()


def has_no_result():
    """Match a Deferred that has not yet fired.

    For example, this will pass::

        assert_that(defer.Deferred(), has_no_result())

    But this will fail:

    >>> assert_that(defer.succeed(None), has_no_result())
    Traceback (most recent call last):
      ...
      File "testtools/assertions.py", line 22, in assert_that
        raise MismatchError(matchee, matcher, mismatch, verbose)
    testtools.matchers._impl.MismatchError: No result expected on <Deferred at ...
        current result: None>, found None instead

    As will this:

    >>> assert_that(defer.fail(RuntimeError('foo')), has_no_result())
    Traceback (most recent call last):
      ...
      File "testtools/assertions.py", line 22, in assert_that
        raise MismatchError(matchee, matcher, mismatch, verbose)
    testtools.matchers._impl.MismatchError: No result expected on <Deferred at ...
        current result: <twisted.python.failure.Failure
        <type 'exceptions.RuntimeError'>>>, found
        <twisted.python.failure.Failure <type 'exceptions.RuntimeError'>> instead
    """
    return _NO_RESULT


class _Succeeded(Matcher[Any]):
    """Matches a Deferred that has fired successfully."""

    def __init__(self, matcher):
        """Construct a ``_Succeeded`` matcher."""
        self._matcher = matcher

    @staticmethod
    def _got_failure(deferred, failure):
        deferred.addErrback(lambda _: None)
        return Mismatch(
            f"Success result expected on {deferred!r}, found failure result "
            f"instead: {failure!r}",
            {"traceback": failure_content(failure)},
        )

    @staticmethod
    def _got_no_result(deferred):
        return Mismatch(
            f"Success result expected on {deferred!r}, found no result instead"
        )

    def match(self, deferred):
        """Match against the successful result of ``deferred``."""
        return on_deferred_result(
            deferred,
            on_success=lambda _, value: self._matcher.match(value),
            on_failure=self._got_failure,
            on_no_result=self._got_no_result,
        )


def succeeded(matcher):
    """Match a Deferred that has fired successfully.

    For example::

        fires_with_the_answer = succeeded(Equals(42))
        deferred = defer.succeed(42)
        assert_that(deferred, fires_with_the_answer)

    This assertion will pass. However, if ``deferred`` had fired with a
    different value, or had failed, or had not fired at all, then it would
    fail.

    Use this instead of
    :py:meth:`twisted.trial.unittest.SynchronousTestCase.successResultOf`.

    :param matcher: A matcher to match against the result of a
        :class:`~twisted.internet.defer.Deferred`.
    :return: A matcher that can be applied to a synchronous
        :class:`~twisted.internet.defer.Deferred`.
    """
    return _Succeeded(matcher)


class _Failed(Matcher[Any]):
    """Matches a Deferred that has failed."""

    def __init__(self, matcher):
        self._matcher = matcher

    def _got_failure(self, deferred, failure):
        # We have handled the failure, so suppress its output.
        deferred.addErrback(lambda _: None)
        return self._matcher.match(failure)

    @staticmethod
    def _got_success(deferred, success):
        return Mismatch(
            f"Failure result expected on {deferred!r}, found success "
            f"result ({success!r}) instead"
        )

    @staticmethod
    def _got_no_result(deferred):
        return Mismatch(
            f"Failure result expected on {deferred!r}, found no result instead"
        )

    def match(self, deferred):
        return on_deferred_result(
            deferred,
            on_success=self._got_success,
            on_failure=self._got_failure,
            on_no_result=self._got_no_result,
        )


def failed(matcher):
    """Match a Deferred that has failed.

    For example::

        error = RuntimeError('foo')
        fails_at_runtime = failed(
            AfterPreprocessing(lambda f: f.value, Equals(error)))
        deferred = defer.fail(error)
        assert_that(deferred, fails_at_runtime)

    This assertion will pass. However, if ``deferred`` had fired successfully,
    had failed with a different error, or had not fired at all, then it would
    fail.

    Use this instead of
    :py:meth:`twisted.trial.unittest.SynchronousTestCase.failureResultOf`.

    :param matcher: A matcher to match against the result of a failing
        :class:`~twisted.internet.defer.Deferred`.
    :return: A matcher that can be applied to a synchronous
        :class:`~twisted.internet.defer.Deferred`.
    """
    return _Failed(matcher)



# Copyright (c) 2008-2012 testtools developers. See LICENSE for details.

from collections.abc import Callable
from typing import Any, ClassVar

from testtools.matchers._impl import Matcher


class TestMatchersInterface:
    """Mixin class that provides test methods for matcher interfaces."""

    __test__ = False  # Tell pytest not to collect this as a test class

    matches_matcher: ClassVar[Matcher[Any]]
    matches_matches: ClassVar[list[Any]]
    matches_mismatches: ClassVar[list[Any]]
    str_examples: ClassVar[Any]
    describe_examples: ClassVar[Any]

    assertEqual: Callable[..., Any]
    assertNotEqual: Callable[..., Any]
    assertThat: Callable[..., Any]

    def test_matches_match(self) -> None:
        matcher = self.matches_matcher
        matches = self.matches_matches
        mismatches = self.matches_mismatches
        for candidate in matches:
            self.assertEqual(None, matcher.match(candidate))
        for candidate in mismatches:
            mismatch = matcher.match(candidate)
            self.assertNotEqual(None, mismatch)
            self.assertNotEqual(None, getattr(mismatch, "describe", None))

    def test__str__(self) -> None:
        # [(expected, object to __str__)].
        from testtools.matchers._doctest import DocTestMatches

        examples = self.str_examples
        for expected, matcher in examples:
            self.assertThat(matcher, DocTestMatches(expected))

    def test_describe_difference(self) -> None:
        # [(expected, matchee, matcher), ...]
        examples = self.describe_examples
        for difference, matchee, matcher in examples:
            mismatch = matcher.match(matchee)
            self.assertEqual(difference, mismatch.describe())

    def test_mismatch_details(self) -> None:
        # The mismatch object must provide get_details, which must return a
        # dictionary mapping names to Content objects.
        examples = self.describe_examples
        for difference, matchee, matcher in examples:
            mismatch = matcher.match(matchee)
            details = mismatch.get_details()
            self.assertEqual(dict(details), details)



# Copyright (c) 2009-2015 testtools developers. See LICENSE for details.

"""Test suites and related things."""

__all__ = [
    "ConcurrentStreamTestSuite",
    "ConcurrentTestSuite",
    "filter_by_ids",
    "iterate_tests",
    "sorted_tests",
]

import sys
import threading
import unittest
from collections import Counter
from collections.abc import Callable, Container, Generator, Iterable
from pprint import pformat
from queue import Queue
from typing import Any, Protocol, TypeAlias, TypeVar, runtime_checkable

import testtools

try:
    import fixtures
except ImportError:
    fixtures = None  # type: ignore

_T = TypeVar("_T")

# Type alias for objects that can be test suites or test cases
TestSuiteOrCase: TypeAlias = unittest.TestSuite | unittest.TestCase


class _Stoppable(Protocol):
    """Protocol for result objects that can be stopped."""

    def stop(self) -> None: ...


class _StreamResultLike(_Stoppable, Protocol):
    """Protocol for stream result objects with test run lifecycle."""

    def startTestRun(self) -> None: ...
    def stopTestRun(self) -> None: ...


@runtime_checkable
class _Sortable(Protocol):
    """Protocol for test suites that can be sorted."""

    def sort_tests(self) -> None: ...


def iterate_tests(
    test_suite_or_case: TestSuiteOrCase,
) -> Generator[unittest.TestCase, None, None]:
    """Iterate through all of the test cases in 'test_suite_or_case'."""
    if isinstance(test_suite_or_case, Iterable) and not isinstance(
        test_suite_or_case, unittest.TestCase
    ):
        for test in test_suite_or_case:
            yield from iterate_tests(test)
    else:
        yield test_suite_or_case


class ConcurrentTestSuite(unittest.TestSuite):
    """A TestSuite whose run() calls out to a concurrency strategy."""

    def __init__(
        self,
        suite: unittest.TestSuite,
        make_tests: Callable[[unittest.TestSuite], Iterable[unittest.TestCase]],
        wrap_result: Callable[[Any, int], unittest.TestResult] | None = None,
    ) -> None:
        """Create a ConcurrentTestSuite to execute suite.

        :param suite: A suite to run concurrently.
        :param make_tests: A helper function to split the tests in the
            ConcurrentTestSuite into some number of concurrently executing
            sub-suites. make_tests must take a suite, and return an iterable
            of TestCase-like object, each of which must have a run(result)
            method.
        :param wrap_result: An optional function that takes a thread-safe
            result and a thread number and must return a ``TestResult``
            object. If not provided, then ``ConcurrentTestSuite`` will just
            use a ``ThreadsafeForwardingResult`` wrapped around the result
            passed to ``run()``.
        """
        super().__init__([suite])
        self.make_tests = make_tests
        self._custom_wrap_result = wrap_result

    def _wrap_result(
        self, thread_safe_result: unittest.TestResult, thread_number: int
    ) -> unittest.TestResult:
        """Wrap a thread-safe result before sending it test results.

        You can either override this in a subclass or pass your own
        ``wrap_result`` in to the constructor.  The latter is preferred.
        """
        if self._custom_wrap_result:
            return self._custom_wrap_result(thread_safe_result, thread_number)
        return thread_safe_result

    def run(
        self, result: unittest.TestResult, debug: bool = False
    ) -> unittest.TestResult:
        """Run the tests concurrently.

        This calls out to the provided make_tests helper, and then serialises
        the results so that result only sees activity from one TestCase at
        a time.

        ConcurrentTestSuite provides no special mechanism to stop the tests
        returned by make_tests, it is up to the make_tests to honour the
        shouldStop attribute on the result object they are run with, which will
        be set if an exception is raised in the thread which
        ConcurrentTestSuite.run is called in.
        """
        tests = self.make_tests(self)
        try:
            threads = {}
            queue: Queue[Any] = Queue()
            semaphore = threading.Semaphore(1)
            for i, test in enumerate(tests):
                process_result = self._wrap_result(
                    testtools.ThreadsafeForwardingResult(result, semaphore), i
                )
                reader_thread = threading.Thread(
                    target=self._run_test, args=(test, process_result, queue)
                )
                threads[test] = reader_thread, process_result
                reader_thread.start()
            while threads:
                finished_test = queue.get()
                threads[finished_test][0].join()
                del threads[finished_test]
        except:
            for thread, process_result in threads.values():
                process_result.stop()
            raise
        return result

    def _run_test(
        self,
        test: unittest.TestCase,
        process_result: unittest.TestResult,
        queue: Queue[unittest.TestCase],
    ) -> None:
        try:
            try:
                test.run(process_result)
            except Exception:
                # The run logic itself failed.
                case = testtools.ErrorHolder("broken-runner", error=sys.exc_info())
                case.run(process_result)
        finally:
            queue.put(test)


class ConcurrentStreamTestSuite:
    """A TestSuite whose run() parallelises."""

    def __init__(
        self, make_tests: Callable[[], Iterable[tuple[TestSuiteOrCase, str | None]]]
    ) -> None:
        """Create a ConcurrentTestSuite to execute tests returned by make_tests.

        :param make_tests: A helper function that should return some number
            of concurrently executable test suite / test case objects.
            make_tests must take no parameters and return an iterable of
            tuples. Each tuple must be of the form (case, route_code), where
            case is a TestCase-like object with a run(result) method, and
            route_code is either None or a unicode string.
        """
        super().__init__()
        self.make_tests = make_tests

    def run(self, result: testtools.StreamResult, debug: bool = False) -> None:
        """Run the tests concurrently.

        This calls out to the provided make_tests helper to determine the
        concurrency to use and to assign routing codes to each worker.

        ConcurrentTestSuite provides no special mechanism to stop the tests
        returned by make_tests, it is up to the made tests to honour the
        shouldStop attribute on the result object they are run with, which will
        be set if the test run is to be aborted.

        The tests are run with an ExtendedToStreamDecorator wrapped around a
        StreamToQueue instance. ConcurrentStreamTestSuite dequeues events from
        the queue and forwards them to result. Tests can therefore be either
        original unittest tests (or compatible tests), or new tests that emit
        StreamResult events directly.

        :param result: A StreamResult instance. The caller is responsible for
            calling startTestRun on this instance prior to invoking suite.run,
            and stopTestRun subsequent to the run method returning.
        """
        tests = self.make_tests()
        try:
            threads = {}
            queue: Queue[Any] = Queue()
            for test, route_code in tests:
                to_queue = testtools.StreamToQueue(queue, route_code)
                process_result = testtools.ExtendedToStreamDecorator(
                    testtools.TimestampingStreamResult(to_queue)
                )
                runner_thread = threading.Thread(
                    target=self._run_test, args=(test, process_result, route_code)
                )
                threads[to_queue] = runner_thread, process_result
                runner_thread.start()
            while threads:
                event_dict = queue.get()
                event = event_dict.pop("event")
                if event == "status":
                    result.status(**event_dict)
                elif event == "stopTestRun":
                    thread = threads.pop(event_dict["result"])[0]
                    thread.join()
                elif event == "startTestRun":
                    pass
                else:
                    raise ValueError(f"unknown event type {event!r}")
        except:
            for thread, process_result in threads.values():
                # Signal to each TestControl in the ExtendedToStreamDecorator
                # that the thread should stop running tests and cleanup
                process_result.stop()
            raise

    def _run_test(
        self,
        test: TestSuiteOrCase,
        process_result: unittest.TestResult,
        route_code: str | None,
    ) -> None:
        process_result.startTestRun()
        try:
            try:
                test.run(process_result)
            except Exception:
                # The run logic itself failed.
                case = testtools.ErrorHolder(
                    f"broken-runner-'{route_code}'", error=sys.exc_info()
                )
                case.run(process_result)
        finally:
            process_result.stopTestRun()


class FixtureSuite(unittest.TestSuite):
    def __init__(
        self, fixture: "fixtures.Fixture", tests: Iterable[unittest.TestCase]
    ) -> None:
        super().__init__(tests)
        self._fixture = fixture

    def run(
        self, result: unittest.TestResult, debug: bool = False
    ) -> unittest.TestResult:
        self._fixture.setUp()
        try:
            return super().run(result, debug)
        finally:
            self._fixture.cleanUp()

    def sort_tests(self) -> None:
        sorted_suite = sorted_tests(self, True)
        self._tests[:] = sorted_suite._tests


def _flatten_tests(
    suite_or_case: TestSuiteOrCase, unpack_outer: bool = False
) -> list[tuple[str | None, TestSuiteOrCase]]:
    if isinstance(suite_or_case, unittest.TestCase):
        # Not iterable, it's a test case.
        return [(suite_or_case.id(), suite_or_case)]

    # It's a suite, try to iterate
    if type(suite_or_case) in (unittest.TestSuite,) or unpack_outer:
        # Plain old test suite (or any others we may add).
        result: list[tuple[str | None, TestSuiteOrCase]] = []
        for test in suite_or_case:
            # Recurse to flatten - test is TestSuiteOrCase from the suite
            result.extend(_flatten_tests(test))
        return result
    else:
        # Find any old actual test and grab its id.
        suite_id: str | None = None
        tests_iter = iterate_tests(suite_or_case)
        for test in tests_iter:
            suite_id = test.id()
            break
        # If it has a sort_tests method, call that.
        if isinstance(suite_or_case, _Sortable):
            suite_or_case.sort_tests()
        return [(suite_id, suite_or_case)]


def filter_by_ids(
    suite_or_case: _T, test_ids: Container[str]
) -> _T | unittest.TestSuite:
    """Remove tests from suite_or_case where their id is not in test_ids.

    :param suite_or_case: A test suite or test case.
    :param test_ids: Something that supports the __contains__ protocol.
    :return: suite_or_case, unless suite_or_case was a case that itself
        fails the predicate when it will return a new unittest.TestSuite with
        no contents.

    This helper exists to provide backwards compatibility with older versions
    of Python (currently all versions :)) that don't have a native
    filter_by_ids() method on Test(Case|Suite).

    For subclasses of TestSuite, filtering is done by:
        - attempting to call suite.filter_by_ids(test_ids)
        - if there is no method, iterating the suite and identifying tests to
          remove, then removing them from _tests, manually recursing into
          each entry.

    For objects with an id() method - TestCases, filtering is done by:
        - attempting to return case.filter_by_ids(test_ids)
        - if there is no such method, checking for case.id() in test_ids
          and returning case if it is, or TestSuite() if it is not.

    For anything else, it is not filtered - it is returned as-is.

    To provide compatibility with this routine for a custom TestSuite, just
    define a filter_by_ids() method that will return a TestSuite equivalent to
    the original minus any tests not in test_ids.
    Similarly to provide compatibility for a custom TestCase that does
    something unusual define filter_by_ids to return a new TestCase object
    that will only run test_ids that are in the provided container. If none
    would run, return an empty TestSuite().

    The contract for this function does not require mutation - each filtered
    object can choose to return a new object with the filtered tests. However
    because existing custom TestSuite classes in the wild do not have this
    method, we need a way to copy their state correctly which is tricky:
    thus the backwards-compatible code paths attempt to mutate in place rather
    than guessing how to reconstruct a new suite.
    """
    # Compatible objects
    if hasattr(suite_or_case, "filter_by_ids"):
        return suite_or_case.filter_by_ids(test_ids)  # type: ignore[no-any-return]
    # TestCase objects.
    if hasattr(suite_or_case, "id"):
        if suite_or_case.id() in test_ids:
            return suite_or_case
        else:
            return unittest.TestSuite()
    # Standard TestSuites or derived classes [assumed to be mutable].
    if isinstance(suite_or_case, unittest.TestSuite):
        filtered: list[TestSuiteOrCase] = []
        for item in suite_or_case:
            filtered.append(filter_by_ids(item, test_ids))
        suite_or_case._tests[:] = filtered  # type: ignore[assignment]
    # Everything else:
    return suite_or_case


def sorted_tests(
    suite_or_case: TestSuiteOrCase, unpack_outer: bool = False
) -> unittest.TestSuite:
    """Sort suite_or_case while preserving non-vanilla TestSuites."""
    # Duplicate test id can induce TypeError in Python 3.3.
    # Detect the duplicate test ids, raise exception when found.
    seen = Counter(case.id() for case in iterate_tests(suite_or_case))
    duplicates = {test_id: count for test_id, count in seen.items() if count > 1}
    if duplicates:
        raise ValueError(f"Duplicate test ids detected: {pformat(duplicates)}")

    tests = _flatten_tests(suite_or_case, unpack_outer=unpack_outer)
    tests.sort()
    return unittest.TestSuite([test for (sort_key, test) in tests])



# Copyright (c) 2008-2012 testtools developers. See LICENSE for details.

"""All the matchers.

Matchers, a way to express complex assertions outside the testcase.

Inspired by 'hamcrest'.

Matcher provides the abstract API that all matchers need to implement.

Bundled matchers are listed in __all__: a list can be obtained by running
$ python -c 'import testtools.matchers; print testtools.matchers.__all__'
"""

__all__ = [
    "AfterPreprocessing",
    "AllMatch",
    "Always",
    "Annotate",
    "AnyMatch",
    "ContainedByDict",
    "Contains",
    "ContainsAll",
    "ContainsDict",
    "DirContains",
    "DirExists",
    "DocTestMatches",
    "EndsWith",
    "Equals",
    "FileContains",
    "FileExists",
    "GreaterThan",
    "GreaterThanOrEqual",
    "HasLength",
    "HasPermissions",
    "Is",
    "IsDeprecated",
    "IsInstance",
    "KeysEqual",
    "LessThan",
    "LessThanOrEqual",
    "MatchesAll",
    "MatchesAny",
    "MatchesDict",
    "MatchesException",
    "MatchesListwise",
    "MatchesPredicate",
    "MatchesPredicateWithParams",
    "MatchesRegex",
    "MatchesSetwise",
    "MatchesStructure",
    "Nearly",
    "Never",
    "Not",
    "NotEquals",
    "PathExists",
    "Raises",
    "SameMembers",
    "SamePath",
    "StartsWith",
    "TarballContains",
    "WarningMessage",
    "Warnings",
    "raises",
]

from ._basic import (
    Contains,
    EndsWith,
    Equals,
    GreaterThan,
    GreaterThanOrEqual,
    HasLength,
    Is,
    IsInstance,
    LessThan,
    LessThanOrEqual,
    MatchesRegex,
    Nearly,
    NotEquals,
    SameMembers,
    StartsWith,
)
from ._const import (
    Always,
    Never,
)
from ._datastructures import (
    ContainsAll,
    MatchesListwise,
    MatchesSetwise,
    MatchesStructure,
)
from ._dict import (
    ContainedByDict,
    ContainsDict,
    KeysEqual,
    MatchesDict,
)
from ._doctest import (
    DocTestMatches,
)
from ._exception import (
    MatchesException,
    Raises,
    raises,
)
from ._filesystem import (
    DirContains,
    DirExists,
    FileContains,
    FileExists,
    HasPermissions,
    PathExists,
    SamePath,
    TarballContains,
)
from ._higherorder import (
    AfterPreprocessing,
    AllMatch,
    Annotate,
    AnyMatch,
    MatchesAll,
    MatchesAny,
    MatchesPredicate,
    MatchesPredicateWithParams,
    Not,
)

# XXX: These are not explicitly included in __all__.  It's unclear how much of
# the public interface they really are.
from ._impl import (  # noqa: F401
    Matcher,
    Mismatch,
    MismatchError,
)
from ._warnings import (
    IsDeprecated,
    WarningMessage,
    Warnings,
)



# Copyright (c) 2026 testtools developers. See LICENSE for details.

from types import TracebackType
from typing import TypeAlias

# Type for exc_info tuples from sys.exc_info()
ExcInfo: TypeAlias = tuple[type[BaseException], BaseException, TracebackType | None]
OptExcInfo: TypeAlias = ExcInfo | tuple[None, None, None]



# Copyright (c) 2010-2012 testtools developers. See LICENSE for details.

from collections.abc import Callable
from typing import Any, TypeVar

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
R = TypeVar("R")


def map_values(function: Callable[[V], R], dictionary: dict[K, V]) -> dict[K, R]:
    """Map ``function`` across the values of ``dictionary``.

    :return: A dict with the same keys as ``dictionary``, where the value
        of each key ``k`` is ``function(dictionary[k])``.
    """
    return {k: function(dictionary[k]) for k in dictionary}


def filter_values(function: Callable[[V], bool], dictionary: dict[K, V]) -> dict[K, V]:
    """Filter ``dictionary`` by its values using ``function``."""
    return {k: v for k, v in dictionary.items() if function(v)}


def dict_subtract(a: dict[K, V], b: dict[K, Any]) -> dict[K, V]:
    """Return the part of ``a`` that's not in ``b``."""
    return {k: a[k] for k in set(a) - set(b)}


def list_subtract(a: list[T], b: list[T]) -> list[T]:
    """Return a list ``a`` without the elements of ``b``.

    If a particular value is in ``a`` twice and ``b`` once then the returned
    list then that value will appear once in the returned list.
    """
    a_only = list(a)
    for x in b:
        if x in a_only:
            a_only.remove(x)
    return a_only



# Copyright (c) 2009-2012 testtools developers. See LICENSE for details.

"""Content - a MIME-like Content object."""

__all__ = [
    "Content",
    "TracebackContent",
    "attach_file",
    "content_from_file",
    "content_from_stream",
    "json_content",
    "text_content",
]

import codecs
import functools
import json
import os
import traceback
import types
from collections.abc import Callable, Iterable, Iterator
from typing import IO, Any, Protocol, TypeAlias, runtime_checkable

from testtools._types import OptExcInfo
from testtools.content_type import JSON, UTF8_TEXT, ContentType

# Type for JSON-serializable data
JSONType: TypeAlias = (
    dict[str, "JSONType"] | list["JSONType"] | str | int | float | bool | None
)


class _Detailed(Protocol):
    """Protocol for objects that have an addDetail method."""

    def addDetail(self, name: str, content_object: "Content") -> None: ...


@runtime_checkable
class _TestCase(Protocol):
    """Protocol for test objects used in TracebackContent."""

    failureException: Any  # Can be type[BaseException], tuple, or None


DEFAULT_CHUNK_SIZE = 4096

STDOUT_LINE = "\nStdout:\n%s"
STDERR_LINE = "\nStderr:\n%s"


def _iter_chunks(
    stream: IO[bytes],
    chunk_size: int,
    seek_offset: int | None = None,
    seek_whence: int = 0,
) -> Iterator[bytes]:
    """Read 'stream' in chunks of 'chunk_size'.

    :param stream: A file-like object to read from.
    :param chunk_size: The size of each read from 'stream'.
    :param seek_offset: If non-None, seek before iterating.
    :param seek_whence: Pass through to the seek call, if seeking.
    """
    if seek_offset is not None:
        stream.seek(seek_offset, seek_whence)
    chunk = stream.read(chunk_size)
    while chunk:
        yield chunk
        chunk = stream.read(chunk_size)


class Content:
    """A MIME-like Content object.

    'Content' objects can be serialised to bytes using the iter_bytes method.
    If the 'Content-Type' is recognised by other code, they are welcome to
    look for richer contents that mere byte serialisation - for example in
    memory object graphs etc. However, such code MUST be prepared to receive
    a generic 'Content' object that has been reconstructed from a byte stream.

    :ivar content_type: The content type of this Content.
    """

    def __init__(
        self, content_type: ContentType, get_bytes: Callable[[], Iterable[bytes]]
    ) -> None:
        """Create a ContentType."""
        if None in (content_type, get_bytes):
            raise ValueError(f"None not permitted in {content_type!r}, {get_bytes!r}")
        self.content_type = content_type
        self._get_bytes = get_bytes

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Content):
            return NotImplemented
        return bool(
            self.content_type == other.content_type
            and b"".join(self.iter_bytes()) == b"".join(other.iter_bytes())
        )

    def as_text(self) -> str:
        """Return all of the content as text.

        This is only valid where ``iter_text`` is.  It will load all of the
        content into memory.  Where this is a concern, use ``iter_text``
        instead.
        """
        return "".join(self.iter_text())

    def iter_bytes(self) -> Iterator[bytes]:
        """Iterate over bytestrings of the serialised content."""
        return iter(self._get_bytes())

    def iter_text(self) -> Iterator[str]:
        """Iterate over the text of the serialised content.

        This is only valid for text MIME types, and will use ISO-8859-1 if
        no charset parameter is present in the MIME type. (This is somewhat
        arbitrary, but consistent with RFC2617 3.7.1).

        :raises ValueError: If the content type is not "text/*".
        """
        if self.content_type.type != "text":
            raise ValueError(f"Not a text type {self.content_type!r}")
        return self._iter_text()

    def _iter_text(self) -> Iterator[str]:
        """Worker for iter_text - does the decoding."""
        encoding = self.content_type.parameters.get("charset", "ISO-8859-1")
        decoder = codecs.getincrementaldecoder(encoding)()
        for bytes in self.iter_bytes():
            yield decoder.decode(bytes)
        final = decoder.decode(b"", True)
        if final:
            yield final

    def __repr__(self) -> str:
        return (
            f"<Content type={self.content_type!r}, "
            f"value={b''.join(self.iter_bytes())!r}>"
        )


class StackLinesContent(Content):
    """Content object for stack lines.

    This adapts a list of "preprocessed" stack lines into a 'Content' object.
    The stack lines are most likely produced from ``traceback.extract_stack``
    or ``traceback.extract_tb``.

    text/x-traceback;language=python is used for the mime type, in order to
    provide room for other languages to format their tracebacks differently.
    """

    # Whether or not to hide layers of the stack trace that are
    # unittest/testtools internal code.  Defaults to True since the
    # system-under-test is rarely unittest or testtools.
    HIDE_INTERNAL_STACK = True

    def __init__(
        self,
        stack_lines: traceback.StackSummary,
        prefix_content: str = "",
        postfix_content: str = "",
    ) -> None:
        """Create a StackLinesContent for ``stack_lines``.

        :param stack_lines: A list of preprocessed stack lines, probably
            obtained by calling ``traceback.extract_stack`` or
            ``traceback.extract_tb``.
        :param prefix_content: If specified, a unicode string to prepend to the
            text content.
        :param postfix_content: If specified, a unicode string to append to the
            text content.
        """
        content_type = ContentType(
            "text", "x-traceback", {"language": "python", "charset": "utf8"}
        )
        value = (
            prefix_content + self._stack_lines_to_unicode(stack_lines) + postfix_content
        )
        super().__init__(content_type, lambda: [value.encode("utf8")])

    def _stack_lines_to_unicode(self, stack_lines: traceback.StackSummary) -> str:
        """Converts a list of pre-processed stack lines into a unicode string."""
        msg_lines = traceback.format_list(stack_lines)
        return "".join(msg_lines)


class TracebackContent(Content):
    """Content object for tracebacks.

    This adapts an exc_info tuple to the 'Content' interface.
    'text/x-traceback;language=python' is used for the mime type, in order to
    provide room for other languages to format their tracebacks differently.
    """

    def __init__(
        self,
        err: OptExcInfo,
        test: _TestCase | None,
        capture_locals: bool = False,
    ) -> None:
        """Create a TracebackContent for ``err``.

        :param err: An exc_info error tuple.
        :param test: A test object used to obtain failureException.
        :param capture_locals: If true, show locals in the traceback.
        """
        if err is None:
            raise ValueError("err may not be None")

        exctype, value, tb = err
        # Ensure we have a real exception, not the (None, None, None) variant
        assert exctype is not None, "exctype must not be None"
        assert value is not None, "value must not be None"
        # Skip test runner traceback levels
        if StackLinesContent.HIDE_INTERNAL_STACK:
            while tb and "__unittest" in tb.tb_frame.f_globals:
                tb = tb.tb_next

        limit = None
        # Disabled due to https://bugs.launchpad.net/testtools/+bug/1188420
        if (
            False
            and StackLinesContent.HIDE_INTERNAL_STACK
            and test is not None
            and test.failureException
            and isinstance(value, test.failureException)
        ):
            # Skip assert*() traceback levels
            limit = 0
            while tb and not self._is_relevant_tb_level(tb):
                limit += 1
                tb = tb.tb_next

        stack_lines = list(
            traceback.TracebackException(
                exctype, value, tb, limit=limit, capture_locals=capture_locals
            ).format()
        )
        content_type = ContentType(
            "text", "x-traceback", {"language": "python", "charset": "utf8"}
        )
        super().__init__(content_type, lambda: [x.encode("utf8") for x in stack_lines])


def StacktraceContent(prefix_content: str = "", postfix_content: str = "") -> Content:
    """Content object for stack traces.

    This function will create and return a 'Content' object that contains a
    stack trace.

    The mime type is set to 'text/x-traceback;language=python', so other
    languages can format their stack traces differently.

    :param prefix_content: A unicode string to add before the stack lines.
    :param postfix_content: A unicode string to add after the stack lines.
    """
    stack = traceback.walk_stack(None)

    def filter_stack(
        stack: Iterator[tuple[types.FrameType, int]],
    ) -> Iterator[tuple[types.FrameType, int]]:
        # Discard the filter_stack frame.
        next(stack)
        # Discard the StacktraceContent frame.
        next(stack)
        for f, f_lineno in stack:
            if StackLinesContent.HIDE_INTERNAL_STACK:
                if "__unittest" in f.f_globals:
                    return
                yield f, f_lineno

    extract = traceback.StackSummary.extract(filter_stack(stack))
    extract.reverse()
    return StackLinesContent(extract, prefix_content, postfix_content)


def json_content(json_data: JSONType) -> Content:
    """Create a JSON Content object from JSON-encodeable data."""
    json_str = json.dumps(json_data)
    # The json module perversely returns native str not bytes
    data = json_str.encode("utf8")
    return Content(JSON, lambda: [data])


def text_content(text: str) -> Content:
    """Create a Content object from some text.

    This is useful for adding details which are short strings.
    """
    if not isinstance(text, str):
        raise TypeError(
            f"text_content must be given text, not '{type(text).__name__}'."
        )
    return Content(UTF8_TEXT, lambda: [text.encode("utf8")])


def maybe_wrap(
    wrapper: Callable[..., Any], func: Callable[..., Any]
) -> Callable[..., Any]:
    """Merge metadata for func into wrapper if functools is present."""
    if functools is not None:
        wrapper = functools.update_wrapper(wrapper, func)
    return wrapper


def content_from_file(
    path: str,
    content_type: ContentType | None = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    buffer_now: bool = False,
    seek_offset: int | None = None,
    seek_whence: int = 0,
) -> Content:
    """Create a Content object from a file on disk.

    Note that unless ``buffer_now`` is explicitly passed in as True, the file
    will only be read from when ``iter_bytes`` is called.

    :param path: The path to the file to be used as content.
    :param content_type: The type of content.  If not specified, defaults
        to UTF8-encoded text/plain.
    :param chunk_size: The size of chunks to read from the file.
        Defaults to ``DEFAULT_CHUNK_SIZE``.
    :param buffer_now: If True, read the file from disk now and keep it in
        memory. Otherwise, only read when the content is serialized.
    :param seek_offset: If non-None, seek within the stream before reading it.
    :param seek_whence: If supplied, pass to ``stream.seek()`` when seeking.
    """
    if content_type is None:
        content_type = UTF8_TEXT

    def reader() -> Iterable[bytes]:
        with open(path, "rb") as stream:
            yield from _iter_chunks(stream, chunk_size, seek_offset, seek_whence)

    return content_from_reader(reader, content_type, buffer_now)


def content_from_stream(
    stream: IO[bytes],
    content_type: ContentType | None = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    buffer_now: bool = False,
    seek_offset: int | None = None,
    seek_whence: int = 0,
) -> Content:
    """Create a Content object from a file-like stream.

    Note that unless ``buffer_now`` is explicitly passed in as True, the stream
    will only be read from when ``iter_bytes`` is called.

    :param stream: A file-like object to read the content from. The stream
        is not closed by this function or the 'Content' object it returns.
    :param content_type: The type of content. If not specified, defaults
        to UTF8-encoded text/plain.
    :param chunk_size: The size of chunks to read from the file.
        Defaults to ``DEFAULT_CHUNK_SIZE``.
    :param buffer_now: If True, reads from the stream right now. Otherwise,
        only reads when the content is serialized. Defaults to False.
    :param seek_offset: If non-None, seek within the stream before reading it.
    :param seek_whence: If supplied, pass to ``stream.seek()`` when seeking.
    """
    if content_type is None:
        content_type = UTF8_TEXT

    def reader() -> Iterator[bytes]:
        return _iter_chunks(stream, chunk_size, seek_offset, seek_whence)

    return content_from_reader(reader, content_type, buffer_now)


def content_from_reader(
    reader: Callable[[], Iterable[bytes]],
    content_type: ContentType | None,
    buffer_now: bool,
) -> Content:
    """Create a Content object that will obtain the content from reader.

    :param reader: A callback to read the content. Should return an iterable of
        bytestrings.
    :param content_type: The content type to create.
    :param buffer_now: If True the reader is evaluated immediately and
        buffered.
    """
    if content_type is None:
        content_type = UTF8_TEXT
    if buffer_now:
        contents = list(reader())

        def buffered_reader() -> Iterable[bytes]:
            return contents

        return Content(content_type, buffered_reader)

    return Content(content_type, reader)


def attach_file(
    detailed: _Detailed,
    path: str,
    name: str | None = None,
    content_type: ContentType | None = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    buffer_now: bool = True,
) -> None:
    """Attach a file to this test as a detail.

    This is a convenience method wrapping around ``addDetail``.

    Note that by default the contents of the file will be read immediately. If
    ``buffer_now`` is False, then the file *must* exist when the test result is
    called with the results of this test, after the test has been torn down.

    :param detailed: An object with details
    :param path: The path to the file to attach.
    :param name: The name to give to the detail for the attached file.
    :param content_type: The content type of the file.  If not provided,
        defaults to UTF8-encoded text/plain.
    :param chunk_size: The size of chunks to read from the file.  Defaults
        to something sensible.
    :param buffer_now: If False the file content is read when the content
        object is evaluated rather than when attach_file is called.
        Note that this may be after any cleanups that obj_with_details has, so
        if the file is a temporary file disabling buffer_now may cause the file
        to be read after it is deleted. To handle those cases, using
        attach_file as a cleanup is recommended because it guarantees a
        sequence for when the attach_file call is made::

            detailed.addCleanup(attach_file, 'foo.txt', detailed)
    """
    if name is None:
        name = os.path.basename(path)
    content_object = content_from_file(path, content_type, chunk_size, buffer_now)
    detailed.addDetail(name, content_object)



# Copyright (c) 2008-2015 testtools developers. See LICENSE for details.

"""Compatibility support - kept for backwards compatibility."""

__all__ = [
    "BytesIO",
    "StringIO",
    "text_repr",
    "unicode_output_stream",
]

import codecs
import io
import sys
import types
import unicodedata
import warnings
from io import BytesIO, StringIO  # for backwards-compat
from typing import IO, Any, NoReturn


def reraise(
    exc_class: type[BaseException],
    exc_obj: BaseException,
    exc_tb: types.TracebackType,
    _marker: Any = object(),
) -> NoReturn:
    """Re-raise an exception received from sys.exc_info() or similar."""
    warnings.warn(
        "This is not necessary in Python 3.",
        DeprecationWarning,
        stacklevel=2,
    )
    raise exc_obj.with_traceback(exc_tb)


def _u(s: str) -> str:
    warnings.warn(
        "This is not necessary in Python 3.",
        DeprecationWarning,
        stacklevel=2,
    )
    return s


def _b(s: str) -> bytes:
    """A byte literal."""
    warnings.warn(
        "This is not necessary in Python 3.",
        DeprecationWarning,
        stacklevel=2,
    )
    return s.encode("latin-1")


def _slow_escape(text: str) -> str:
    """Escape unicode ``text`` leaving printable characters unmodified

    The behaviour emulates the Python 3 implementation of repr, see
    unicode_repr in unicodeobject.c and isprintable definition.

    Because this iterates over the input a codepoint at a time, it's slow, and
    does not handle astral characters correctly on Python builds with 16 bit
    rather than 32 bit unicode type.
    """
    output: list[str | bytes] = []
    for c in text:
        o = ord(c)
        if o < 256:
            if o < 32 or 126 < o < 161:
                output.append(c.encode("unicode-escape"))
            elif o == 92:
                # Separate due to bug in unicode-escape codec in Python 2.4
                output.append("\\\\")
            else:
                output.append(c)
        else:
            # To get correct behaviour would need to pair up surrogates here
            if unicodedata.category(c)[0] in "CZ":
                output.append(c.encode("unicode-escape"))
            else:
                output.append(c)
    return "".join(output)  # type: ignore[arg-type]


def text_repr(text: str | bytes, multiline: bool | None = None) -> str:
    """Rich repr for ``text`` returning unicode, triple quoted if ``multiline``."""
    nl = (isinstance(text, bytes) and bytes((0xA,))) or "\n"
    if multiline is None:
        multiline = nl in text  # type: ignore[operator]
    if not multiline:
        # Use normal repr for single line of unicode
        return repr(text)
    prefix = repr(text[:0])[:-2]
    if multiline:
        # To escape multiline strings, split and process each line in turn,
        # making sure that quotes are not escaped.
        offset = len(prefix) + 1
        lines = []
        for line in text.split(nl):  # type: ignore[arg-type]
            r = repr(line)
            q = r[-1]
            lines.append(r[offset:-1].replace("\\" + q, q))
        # Combine the escaped lines and append two of the closing quotes,
        # then iterate over the result to escape triple quotes correctly.
        _semi_done = "\n".join(lines) + "''"
        p = 0
        while True:
            p = _semi_done.find("'''", p)
            if p == -1:
                break
            _semi_done = "\\".join([_semi_done[:p], _semi_done[p:]])
            p += 2
        return "".join([prefix, "'''\\\n", _semi_done, "'"])
    escaped_text = _slow_escape(text)
    # Determine which quote character to use and if one gets prefixed with a
    # backslash following the same logic Python uses for repr() on strings
    quote = "'"
    if "'" in text:
        if '"' in text:
            escaped_text = escaped_text.replace("'", "\\'")
        else:
            quote = '"'
    return "".join([prefix, quote, escaped_text, quote])


def unicode_output_stream(stream: IO[str]) -> IO[str]:
    """Get wrapper for given stream that writes any unicode without exception

    Characters that can't be coerced to the encoding of the stream, or 'ascii'
    if valid encoding is not found, will be replaced. The original stream may
    be returned in situations where a wrapper is determined unneeded.

    The wrapper only allows unicode to be written, not non-ascii bytestrings,
    which is a good thing to ensure sanity and sanitation.
    """
    warnings.warn(
        "This is not necessary in Python 3.",
        DeprecationWarning,
        stacklevel=2,
    )

    if sys.platform == "cli" or isinstance(stream, (io.TextIOWrapper, io.StringIO)):
        # Best to never encode before writing in IronPython, or if it is
        # already a TextIO [which in the io library has no encoding
        # attribute).
        return stream

    try:
        writer = codecs.getwriter(stream.encoding or "")  # type: ignore[attr-defined]
    except (AttributeError, LookupError):
        return codecs.getwriter("ascii")(stream, "replace")  # type: ignore[arg-type, return-value]

    if writer.__module__.rsplit(".", 1)[1].startswith("utf"):
        # The current stream has a unicode encoding so no error handler is needed
        return stream

    # Python 3 doesn't seem to make this easy, handle a common case
    try:
        return stream.__class__(  # type: ignore[call-arg]
            stream.buffer,  # type: ignore[attr-defined]
            stream.encoding,  # type: ignore[attr-defined]
            "replace",
            stream.newlines,  # type: ignore[attr-defined]
            stream.line_buffering,  # type: ignore[attr-defined]
        )
    except AttributeError:
        pass

    return writer(stream, "replace")  # type: ignore[arg-type, return-value]



# Copyright (c) 2009-2012 testtools developers. See LICENSE for details.

"""Test result objects."""

__all__ = [
    "CopyStreamResult",
    "DetailsDict",
    "ExtendedToOriginalDecorator",
    "ExtendedToStreamDecorator",
    "MultiTestResult",
    "ResourcedToStreamDecorator",
    "StreamFailFast",
    "StreamResult",
    "StreamResultRouter",
    "StreamSummary",
    "StreamTagger",
    "StreamToDict",
    "StreamToExtendedDecorator",
    "StreamToQueue",
    "Tagger",
    "TestByTestResult",
    "TestControl",
    "TestResult",
    "TestResultDecorator",
    "TextTestResult",
    "ThreadsafeForwardingResult",
    "TimestampingStreamResult",
]

from testtools.testresult.real import (
    CopyStreamResult,
    DetailsDict,
    ExtendedToOriginalDecorator,
    ExtendedToStreamDecorator,
    MultiTestResult,
    ResourcedToStreamDecorator,
    StreamFailFast,
    StreamResult,
    StreamResultRouter,
    StreamSummary,
    StreamTagger,
    StreamToDict,
    StreamToExtendedDecorator,
    StreamToQueue,
    Tagger,
    TestByTestResult,
    TestControl,
    TestResult,
    TestResultDecorator,
    TextTestResult,
    ThreadsafeForwardingResult,
    TimestampingStreamResult,
)



# Copyright (c) 2016 testtools developers. See LICENSE for details.

__all__ = [
    "Always",
    "Never",
]


from ._impl import Matcher, Mismatch


class _Always(Matcher[object]):
    """Always matches."""

    def __str__(self) -> str:
        return "Always()"

    def match(self, value: object) -> None:
        return None


def Always() -> _Always:
    """Always match.

    That is::

        self.assertThat(x, Always())

    Will always match and never fail, no matter what ``x`` is. Most useful when
    passed to other higher-order matchers (e.g.
    :py:class:`~testtools.matchers.MatchesListwise`).
    """
    return _Always()


class _Never(Matcher[object]):
    """Never matches."""

    def __str__(self) -> str:
        return "Never()"

    def match(self, value: object) -> Mismatch:
        return Mismatch(f"Inevitable mismatch on {value!r}")


def Never() -> _Never:
    """Never match.

    That is::

        self.assertThat(x, Never())

    Will never match and always fail, no matter what ``x`` is. Included for
    completeness with :py:func:`.Always`, but if you find a use for this, let
    us know!
    """
    return _Never()



# Copyright (c) 2009-2016 testtools developers. See LICENSE for details.

"""Doubles of test result objects, useful for testing unittest code."""

import datetime
import unittest
from collections import namedtuple
from collections.abc import Iterable
from typing import Literal, TypeAlias

from testtools._types import OptExcInfo
from testtools.tags import TagContext

__all__ = [
    "ExtendedTestResult",
    "Python3TestResult",
    "StreamResult",
    "TwistedTestResult",
]

# Convenience namedtuple for status events - defined early for use in LogEvent
_StatusEvent = namedtuple(
    "_StatusEvent",
    [
        "name",
        "test_id",
        "test_status",
        "test_tags",
        "runnable",
        "file_name",
        "file_bytes",
        "eof",
        "mime_type",
        "route_code",
        "timestamp",
    ],
)

# Event type aliases using plain tuples with Literal for event names
# This provides type safety while working with the existing plain tuple code
LogEvent: TypeAlias = (
    tuple[Literal["startTestRun"]]
    | tuple[Literal["stopTestRun"]]
    | tuple[Literal["startTest"], unittest.TestCase]
    | tuple[Literal["stopTest"], unittest.TestCase]
    | tuple[Literal["addSuccess"], unittest.TestCase]
    | tuple[Literal["addSuccess"], unittest.TestCase, dict[str, object]]
    | tuple[
        Literal["addError"],
        unittest.TestCase,
        tuple[type, Exception, object] | dict[str, object] | None | object,
    ]
    | tuple[
        Literal["addFailure"],
        unittest.TestCase,
        tuple[type, Exception, object] | dict[str, object] | None | object,
    ]
    | tuple[
        Literal["addExpectedFailure"],
        unittest.TestCase,
        tuple[type, Exception, object] | dict[str, object] | None | object,
    ]
    | tuple[
        Literal["addSkip"],
        unittest.TestCase,
        str | dict[str, object] | None,
    ]
    | tuple[Literal["addUnexpectedSuccess"], unittest.TestCase]
    | tuple[Literal["addUnexpectedSuccess"], unittest.TestCase, dict[str, object]]
    | tuple[Literal["addDuration"], unittest.TestCase, float]
    | tuple[Literal["progress"], int, int]
    | tuple[Literal["tags"], Iterable[str], Iterable[str]]
    | tuple[Literal["time"], datetime.datetime]
    | _StatusEvent
)


class LoggingBase:
    """Basic support for logging of results."""

    def __init__(self, event_log: list[LogEvent] | None = None) -> None:
        if event_log is None:
            event_log = []
        self._events = event_log


class Python3TestResult(LoggingBase):
    """A precisely python 3 like test result, that logs."""

    def __init__(self, event_log: list[LogEvent] | None = None) -> None:
        super().__init__(event_log=event_log)
        self.shouldStop = False
        self._was_successful = True
        self.testsRun = 0
        self.failfast = False
        self.collectedDurations: list[tuple[unittest.TestCase, float]] = []

    def addError(self, test: unittest.TestCase, err: OptExcInfo) -> None:
        self._was_successful = False
        self._events.append(("addError", test, err))
        if self.failfast:
            self.stop()

    def addFailure(self, test: unittest.TestCase, err: OptExcInfo) -> None:
        self._was_successful = False
        self._events.append(("addFailure", test, err))
        if self.failfast:
            self.stop()

    def addSuccess(self, test: unittest.TestCase) -> None:
        self._events.append(("addSuccess", test))

    def addExpectedFailure(self, test: unittest.TestCase, err: OptExcInfo) -> None:
        self._events.append(("addExpectedFailure", test, err))

    def addSkip(self, test: unittest.TestCase, reason: str) -> None:
        self._events.append(("addSkip", test, reason))

    def addUnexpectedSuccess(self, test: unittest.TestCase) -> None:
        self._events.append(("addUnexpectedSuccess", test))
        if self.failfast:
            self.stop()

    def addDuration(self, test: unittest.TestCase, duration: float) -> None:
        self._events.append(("addDuration", test, duration))
        self.collectedDurations.append((test, duration))

    def startTest(self, test: unittest.TestCase) -> None:
        self._events.append(("startTest", test))
        self.testsRun += 1

    def startTestRun(self) -> None:
        self._events.append(("startTestRun",))

    def stop(self) -> None:
        self.shouldStop = True

    def stopTest(self, test: unittest.TestCase) -> None:
        self._events.append(("stopTest", test))

    def stopTestRun(self) -> None:
        self._events.append(("stopTestRun",))

    def wasSuccessful(self) -> bool:
        return self._was_successful


class ExtendedTestResult(Python3TestResult):
    """A test result like the proposed extended unittest result API."""

    def __init__(self, event_log: list[LogEvent] | None = None) -> None:
        super().__init__(event_log)
        self._tags = TagContext()

    def addError(
        self,
        test: unittest.TestCase,
        err: OptExcInfo | None = None,
        details: dict[str, object] | None = None,
    ) -> None:
        self._was_successful = False
        self._events.append(("addError", test, err or details))

    def addFailure(
        self,
        test: unittest.TestCase,
        err: OptExcInfo | None = None,
        details: dict[str, object] | None = None,
    ) -> None:
        self._was_successful = False
        self._events.append(("addFailure", test, err or details))

    def addExpectedFailure(
        self,
        test: unittest.TestCase,
        err: OptExcInfo | None = None,
        details: dict[str, object] | None = None,
    ) -> None:
        self._events.append(("addExpectedFailure", test, err or details))

    def addSkip(
        self,
        test: unittest.TestCase,
        reason: str | None = None,
        details: dict[str, object] | None = None,
    ) -> None:
        self._events.append(("addSkip", test, reason or details))

    def addSuccess(
        self, test: unittest.TestCase, details: dict[str, object] | None = None
    ) -> None:
        if details:
            self._events.append(("addSuccess", test, details))
        else:
            self._events.append(("addSuccess", test))

    def addUnexpectedSuccess(
        self, test: unittest.TestCase, details: dict[str, object] | None = None
    ) -> None:
        self._was_successful = False
        if details is not None:
            self._events.append(("addUnexpectedSuccess", test, details))
        else:
            self._events.append(("addUnexpectedSuccess", test))

    def addDuration(self, test: unittest.TestCase, duration: float) -> None:
        self._events.append(("addDuration", test, duration))

    def progress(self, offset: int, whence: int) -> None:
        self._events.append(("progress", offset, whence))

    def startTestRun(self) -> None:
        super().startTestRun()
        self._was_successful = True
        self._tags = TagContext()

    def startTest(self, test: unittest.TestCase) -> None:
        super().startTest(test)
        self._tags = TagContext(self._tags)

    def stopTest(self, test: unittest.TestCase) -> None:
        # NOTE: In Python 3.12.1 skipped tests may not call startTest()
        if self._tags is not None and self._tags.parent is not None:
            self._tags = self._tags.parent
        super().stopTest(test)

    @property
    def current_tags(self) -> set[str]:
        return self._tags.get_current_tags()

    def tags(self, new_tags: Iterable[str], gone_tags: Iterable[str]) -> None:
        self._tags.change_tags(new_tags, gone_tags)
        self._events.append(("tags", new_tags, gone_tags))

    def time(self, time: datetime.datetime) -> None:
        self._events.append(("time", time))

    def wasSuccessful(self) -> bool:
        return self._was_successful


class TwistedTestResult(LoggingBase):
    """Emulate the relevant bits of :py:class:`twisted.trial.itrial.IReporter`.

    Used to ensure that we can use ``trial`` as a test runner.
    """

    def __init__(self, event_log: list[LogEvent] | None = None) -> None:
        super().__init__(event_log=event_log)
        self._was_successful = True
        self.testsRun = 0

    def startTest(self, test: unittest.TestCase) -> None:
        self.testsRun += 1
        self._events.append(("startTest", test))

    def stopTest(self, test: unittest.TestCase) -> None:
        self._events.append(("stopTest", test))

    def addSuccess(self, test: unittest.TestCase) -> None:
        self._events.append(("addSuccess", test))

    def addError(self, test: unittest.TestCase, error: object) -> None:
        self._was_successful = False
        self._events.append(("addError", test, error))

    def addFailure(self, test: unittest.TestCase, error: object) -> None:
        self._was_successful = False
        self._events.append(("addFailure", test, error))

    def addExpectedFailure(
        self, test: unittest.TestCase, failure: object, todo: object | None = None
    ) -> None:
        self._events.append(("addExpectedFailure", test, failure))

    def addUnexpectedSuccess(
        self, test: unittest.TestCase, todo: object | None = None
    ) -> None:
        self._events.append(("addUnexpectedSuccess", test))

    def addSkip(self, test: unittest.TestCase, reason: str) -> None:
        self._events.append(("addSkip", test, reason))

    def wasSuccessful(self) -> bool:
        return self._was_successful

    def done(self) -> None:
        pass


class StreamResult(LoggingBase):
    """A StreamResult implementation for testing.

    All events are logged to _events.
    """

    def startTestRun(self) -> None:
        self._events.append(("startTestRun",))

    def stopTestRun(self) -> None:
        self._events.append(("stopTestRun",))

    def status(
        self,
        test_id: str | None = None,
        test_status: str | None = None,
        test_tags: set[str] | None = None,
        runnable: bool = True,
        file_name: str | None = None,
        file_bytes: bytes | None = None,
        eof: bool = False,
        mime_type: str | None = None,
        route_code: str | None = None,
        timestamp: datetime.datetime | None = None,
    ) -> None:
        self._events.append(
            _StatusEvent(
                "status",
                test_id,
                test_status,
                test_tags,
                runnable,
                file_name,
                file_bytes,
                eof,
                mime_type,
                route_code,
                timestamp,
            )
        )



# Copyright (c) 2008-2012 testtools developers. See LICENSE for details.

import warnings

from testtools.matchers.test import TestMatchersInterface

warnings.warn(
    "This module is deprecated for removal",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = [
    "TestMatchersInterface",
]



# Copyright (c) 2009-2012 testtools developers. See LICENSE for details.

"""ContentType - a MIME Content Type."""


class ContentType:
    """A content type from http://www.iana.org/assignments/media-types/

    :ivar type: The primary type, e.g. "text" or "application"
    :ivar subtype: The subtype, e.g. "plain" or "octet-stream"
    :ivar parameters: A dict of additional parameters specific to the
        content type.
    """

    type: str
    subtype: str
    parameters: dict[str, str]

    def __init__(
        self, primary_type: str, sub_type: str, parameters: dict[str, str] | None = None
    ) -> None:
        """Create a ContentType."""
        if None in (primary_type, sub_type):
            raise ValueError(f"None not permitted in {primary_type!r}, {sub_type!r}")
        self.type = primary_type
        self.subtype = sub_type
        self.parameters = parameters or {}

    def __eq__(self, other: object) -> bool:
        if type(other) is not ContentType:
            return False
        return self.__dict__ == other.__dict__

    def __repr__(self) -> str:
        if self.parameters:
            params = "; "
            params += "; ".join(
                sorted(f'{k}="{v}"' for k, v in self.parameters.items())
            )
        else:
            params = ""
        return f"{self.type}/{self.subtype}{params}"


JSON = ContentType("application", "json")

UTF8_TEXT = ContentType("text", "plain", {"charset": "utf8"})
